#!/usr/bin/python
"Parses OpenArena/Quake3 logs and writes game statistics to HTML files."

#   pyqscore. Parse OpenArena/Quake3 logs and write statistics to HTML.
#   Copyright (C) 2011  Jose Rodriguez
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, version 2.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.

__author__    = "Jose Rodriguez"
__version__   = "0.91"
__date__      = "19-04-2011"
__license__   = "GPLv2"
__copyright__ = "Copyright (C) 2011  Jose Rodriguez"


import sys
import os
import re
import json
import cPickle
import webbrowser

from operator import mod
from datetime import timedelta, datetime
from random import randint
from pdb import set_trace


###################         OPTIONS      ##############################

MAXPLAYERS = 124
# Maximum number of players displayed in HTML output

SORT_OPTION = 'frags'
# How to sort table columns. Options: deaths, frag_death_ratio,
# frags, games, ping, time, won, won_percentage

BAN_LIST = [ 'adolf hitler','UnnamedPlayer','Beret','Kyonshi','Sergei','Rai','Ayumi','Arachna','Gargoyle','Tony','Angelyss','Tanisha','Dark','Skelebot','Sarge','S_Marine','Jenna','Penguin','Major','Nekoyss','Liz','Assassin','Grism','Merman','Grunt' ]
# Comma-separated list containing the nicks of undesired players.
# Nicks must include the colour codes and be inside quotes. 

MINPLAY = 0
# From 0 to 1, minimum fraction of time a player has to play in a game
# relative to that of the player who played for longer in order appear
# in the statistics. Example: Player A is the first player joining a
# match. The match lasts for 10 minutes. If MINPLAY is 0.7, only the
# statistics of players who joined during the first 3 minutes will count.

NUMBER_OF_QUOTES = 0
# Number of random quotes displayed

OPEN_BROWSER = 'no'
# Open browser when finished by setting this to 'yes'

DUMP_DATA = ''
# Dump processed data to file in JSON format by setting this to 'yes'

GTYPE_OVERRIDE = ''
# If you have a mixed log with ctf and non-ctf games, and you wish
# to display the CTF table but pyqscore doesn't cooperate, set this
# to 'CTF'. Conversely, if you don't want the CTF table, set this
# to 'DM'.

#######################################################################



class Game:
    '''Class with no methods used to store game data.'''
    def __init__(self,number):
        self.number = number            # game number
        self.mapname  = []
        self.players  = {}              # nick: (ping, position)   
        self.pid      = {}              # player id
        self.handicap = {}
        self.teams    = {}
        self.scores   = []
        self.awards   = {}
        self.itemsp   = {}
        self.killsp   = {}             # Stores world frags and suicides
        self.deathsp  = {}             # Stores number of deaths caused by other players
        self.ctf      = {}             # 0: flag taken; 1: capture, 2: flag return; 2: flag fragged
        self.killsp['<world>'] = []
        self.ptime    = {}             # Player time
        self.time     = 0              # Game time 
        self.validp   = []             # Valid game flag
        self.quotes   = set()
        self.weapons  = {}
        


class Server:
    '''Another class to store server data.'''
    def __init__(self):
        #self.name  = server_name
        self.time  = 0
        self.frags = 0
        self.gtype = 0



def totime(string):
    '''Convert strings of the format mmm:ss to an int of seconds'''
    mins, secs = string.split(':')
    time = timedelta(minutes = int(mins), seconds = int(secs)).seconds
    return time



# The most basic error checking conceivable
if  len(sys.argv) < 2:
    print '\nSpecify log file to be processed.\n'
    quit()



log_file = sys.argv[1]
cache_file = str(log_file[:-4]) + '_cache.p'


try:
    file_in = open(log_file, 'r')            
    log     = {}                        # Dictionary to store lines from the log
except(IOError):
    print '\nLog file not found.\n'
    quit()

# If a cache file (a pickled python list) from a previous run exists, use it.
# If present, the cache stores at position [-1] the size of the log
# file from the previous run; at [-2] the number of lines read; at [-3]
# a Server() instance with accumulated server data; at [-4] the
# accumulated list of unique quotes; and at [0:-4] accumulated player data.

try:
    cache = cPickle.load(open(cache_file, 'rb'))
    
    if os.path.getsize(log_file) < cache[-1][1]:
    # If log file's smaller than before, assume it has been
    # overwritten and don't use the cache.
        print '\nLog file size is smaller than the cached one. Processing the entire log file.\n'
        cache_present = 'no'
        key = 1                         # Counter to keep track of previously processed lines.
                                        # If cache does not exist, this number is one.
    else:
        key = cache[-2][1]              # Number of lines read stored at position [-2]
        print '\nCache file found.\n' + str(key) + ' lines already processed.'
        cache_present = 'yes'

except(IOError):
    print '\nNo cache file found. Processing the entire log file.'
    key = 1
    cache_present = 'no'


k     = 1
k_new = 0
for line in file_in:                    # Read log line by line, and
    if k < key:                         # ignore lines read in previous runs
        pass
    else:
        log[k] = line
        k_new += 1
    k += 1

print  '\n' + str(k_new) + ' new lines read.\n'
file_in.close()


if cache_present == 'yes':              
    i = key                             # Overall line counter for main loop
else: 
    cgames = []
    i = 1


server = Server()
cgames = []                             # List of games: instances of class Game
n = 1                                   # Game number
valid_game = False                      # Only completed games will be counted



while i < k:                            # Main loop to process the log

    if log[i].find(' InitGame: ') > 0 and log[i+1].find(' Warmup:') == -1:
    # New game started (no warmup). Begin to parse stuff
        game   = Game(n)                    # New game instance
        pos    = 1                          # Player position in score. Starts at 1
        
        #  0:00 InitGame: \capturelimit\0\g_maxGameClients\0\sv_maxclients\8\timelimit\0\fraglimit\20\dmflags\0\sv_hostname\noname\sv_minRate\0\sv_maxRate\0\sv_minPing\0\sv_maxPing\0\sv_floodProtect\1\sv_allowDownload\1\version\ioQ3 1.33+oa linux-i386 Jul  7 2007\g_gametype\0\protocol\68\mapname\foxhill\sv_privateClients\0\gamename\baseoa\g_needpass\0
        
        #  0:00 InitGame: \dmflags\0\fraglimit\20\timelimit\12\g_gametype\0\sv_privateClients\6\sv_hostname\^1MECAGO^3EN^2DIOS^5!!!! \sv_maxclients\4\sv_minRate\0\sv_maxRate\25000\sv_minPing\0\sv_maxPing\500\sv_floodProtect\1\sv_allowDownload\1\sv_dlURL\http://server/path\g_maxGameClients\22\capturelimit\8\g_delagHitscan\1\g_obeliskRespawnDelay\10\elimination_roundtime\90\elimination_ctf_oneway\0\version\ioq3+oa 1.35 linux-i386 Oct 20 2008\protocol\71\mapname\13base\.Admin\My name\.e-mail\My email\.Location\My location\.OS\My OS\gamename\baseoa\g_needpass\0\g_rockets\0\g_instantgib\0\g_humanplayers\0
        
        #  0:00 InitGame: \g_delagHitscan\1\sv_hostname\noname\sv_minRate\0\sv_maxRate\0\sv_minPing\0\sv_maxPing\0\sv_floodProtect\1\dmflags\0\fraglimit\20\timelimit\0\sv_maxclients\6\g_maxGameClients\0\capturelimit\0\g_allowVote\1\g_voteGametypes\/0/1/3/4/5/6/7/8/9/10/11/12/\g_voteMaxTimelimit\0\g_voteMinTimelimit\0\g_voteMaxFraglimit\0\g_voteMinFraglimit\0\elimination_roundtime\120\g_lms_mode\0\videoflags\7\g_doWarmup\0\version\ioQ3 1.33+oa linux-i386 Oct 22 2008\g_gametype\0\protocol\71\mapname\ce1m7\sv_privateClients\0\sv_allowDownload\0\g_instantgib\0\g_rockets\0\gamename\baseoa\elimflags\0\voteflags\0\g_needpass\0\g_obeliskRespawnDelay\10\g_enableDust\0\g_enableBreath\0\g_altExcellent\0


        regex        = re.compile('mapname[\\\\]([\w]*)')
        mapname      = regex.search(log[i]).group(1)
        game.mapname = mapname
        
        idx          = log[i].find('sv_hostname') + 11
        hostname     = log[i][idx:idx+50].split('\\')[1] # Does this always work?
        server.hostname = hostname                       # I hope so anyway
        
        idx           = log[i].find('g_gametype')
        game.gametype = log[i][idx+11]
        try:
            server.gtype  = int(game.gametype)
        except(ValueError):
            server.gtype = 0                # Default to DM if bad things happen
        
        j = i + 1                           # In-game line counter
                                            # Inner loop packs up stuff from 1 game

        while  j < k:
            this_line = log[j]
            # Process more frequent lines first: Items >> Kill > Userinfo > Awards
            if this_line.find(' Item: ') > 0:
            #  0:35 Item: 1 ammo_lightning
            #100:22 Item: 0 item_health
				pass
                # I don't need items at the moment, so pass and save a lot of time.
				# If they are needed the following lines provide everything required
				# to keep track of the items collected by each player.

                #parts  = this_line[13:].split(' ')
                #client = parts[0]
                #item   = parts[1][:-1]
                ## try statement to avoid _very rare_ cases of damaged logs
                #try:
                    ## Items assigned to player who currently owns specified id.
                    ## In case of client disconnection this may give an erroneous count,
                    ## a circumstance minimised by only storing players with a minimum
                    ## playing time (see below game.validp)
                    #game.itemsp[game.pid[client]].append(item)
                #except:
                    #pass
                    

            elif this_line.find(' Kill: ') > 0:
            #  3:20 Kill: 3 2 10: Gargoyle killed Gargoyle by MOD_RAILGUN
            #100:04 Kill: 0 1 11: ^4kernel panic killed Kyonshi by MOD_LIGHTNING
                k_idx  = this_line.find(' killed ')
                # If somebody's nick contains the string ' killed ' we're screwed
                
                regex  = re.compile('\d:[\s](.*)')            # Fragger's nick
                killer = regex.search(this_line[17:k_idx]).group(1)
                
                d_idx  = k_idx + 6
                b_idx  = this_line.rfind(' b')
                
                killed = this_line[d_idx + 2:b_idx]           # Victim
                weapon = this_line[b_idx + 7 + 1:-1]          # Weapon
                # try statement needed to avoid rare cases of damaged logs:
                # We're looking stuff up on a dictionary, so if the line's broken
                # the key may not exist and python rightfully complains.
                try:
                    if killer == killed:
                        game.killsp[killer].append(weapon)
                    elif killer != '<world>':
                        game.weapons[killer][weapon[0:3]] = game.weapons[killer][weapon[0:3]] + 1
                    else:
                        game.killsp['<world>'].append(killed)
                    game.deathsp[killed] = game.deathsp[killed] + 1
                except:
                    pass
               
                server.frags += 1
              
              
            elif this_line.find(' CTF: ') > 0:
            #  9:40 CTF: 1 1 3: Inhakitor fragged RED's flag carrier!
            # 10:18 CTF: 3 1 0: Mynard Killman got the RED flag!
                p_id  = this_line[12]                         # Player ID
                #team  = this_line[14]                        # Useless datum
                event = this_line[16]              
                # 0: flag taken; 1: flag cap; 2: flag return; 3: flag carrier fragged
                try:
                    game.ctf[game.pid[p_id]][event] = game.ctf[game.pid[p_id]][event] + 1
                except:
                    pass


            elif this_line.find(' Award: ') > 0:
            #  3:02 Award: 4 2: Grunt gained the IMPRESSIVE award!
            # 11:02 Award: 2 1: Kyonshi gained the EXCELLENT award!
                g_idx = this_line.find(' gained ')
                regex = re.compile('\d:\s(\S*\s?\S*)')        # Player name
                result = regex.search(this_line[0:g_idx])
                # Assist, Capture, Defence, Impressive, Excellent 
                name, award = [result.group(1), this_line[g_idx+12:g_idx+13]]
                try:
                    game.awards[name][award] = game.awards[name][award] + 1
                except:
                    pass


            elif this_line.find('UserinfoChanged') > 0:
            #  0:05 ClientUserinfoChanged: 0 n\kernel\t\3\model\sarge/classic\hmodel\sarge/classic\g_redteam\\g_blueteam\\c1\3\c2\5\hc\100\w\0\l\0\tt\0\tl\0
            #103:22 ClientUserinfoChanged: 1 n\Kyonshi\t\0\model\kyonshi\hmodel\kyonshi\c1\4\c2\5\hc\100\w\0\l\0\skill\    5.00\tt\0\tl\0
                regex    = re.compile('Changed:[\s]([\d]*)')      # Retrieve client id
                new_id   = regex.search(this_line).group(1)
                regex    = re.compile('n\\\\([^\\\\]*)')          # retrieve client name
                new_name = regex.search(this_line).group(1)
                regex    = re.compile('\\\\hc\\\\(\d*)')          # Retrieve handicap
                handicap = regex.search(this_line).group(1)
                regex    = re.compile('\\\\t\\\\(\d)')            # Retrieve team. 0: free4all
                team     = regex.search(this_line).group(1)       # 1: red; 2: blue; 3: spect

                if new_name not in game.pid.values():
                    # Initialize dictionaries for new player
                    game.itemsp[new_name]   = []
                    game.killsp[new_name]   = []
                    game.deathsp[new_name]  = 0
                    game.awards[new_name]   = {'A': 0, 'C': 0, 'D': 0, 'E': 0, 'I': 0 }
                    game.handicap[new_name] = handicap
                    game.teams[new_name]    = team
                    game.ctf[new_name]      = {'0': 0, '1': 0, '2': 0, '3': 0}
                    game.weapons[new_name]  = {'SHO': 0, 'GAU': 0, 'MAC': 0, 'GRE': 0, 'ROC': 0,\
                    'PLA': 0, 'RAI': 0, 'LIG': 0, 'BFG': 0, 'TEL': 0,\
                    'NAI': 0, 'CHA': 0 }
                    
                    c_idx = this_line.find('ClientU')
                    game.ptime[new_name]    = totime(this_line[0:c_idx])
                # Keep track of player's current id
                game.pid[new_id] = new_name


            elif this_line.find(' say:') > 0:
                #  2:03 say: ^2ONAK: joder otra vez no
                name = this_line.split(':')[2]
                bs   = this_line.split(':')[3][0:-1]
                game.quotes.add( (name,bs) )


            elif this_line.find(' score: ') > 0:
            #  5:40 score: 6  ping: 85  client: 2 Iagoi
            # 10:14 score: 12  ping: 62  client: 2 Iagoi
                regex = re.compile('(\s?\s?\s? \S*) [\s][^\s]*[\s] (\S?\d*) [\s]+[^\s]*[\s] (\d*) [\s]+[^\s]*[\s] (\d*) \s (.*)', re.VERBOSE)
                result = regex.search(this_line)
                
                [time, score, ping, client, nick] = [result.group(1), result.group(2), result.group(3), result.group(4), result.group(5)]

                game.scores.append([time, score, ping, client, nick])
                game.players[nick] = (ping, pos)
                pos += 1                            # Increase position for next player
                # Players are considered 'valid' if time played is greater than a %
                # of the time played by the 1st player who joined the game.
                # This a) minimises the possibility of wrong item assignment due to 
                # multiple connections/disconnections; b) results in fairer statistics
                if (game.time - game.ptime[nick]) > MINPLAY*(game.time - min(game.ptime.values())):
                    game.validp.append(nick)


            elif this_line.find(' red:') > 0:
            # 20:33 red:4  blue:5
                game.ctfscores = (this_line[11], this_line[19])


            elif this_line.find('Exit: Timelimit hit') > 0 or      \
                 this_line.find('Exit: Fraglimit hit') > 0 or      \
                 this_line.find('Exit: Capturelimit hit') > 0:
            # Game completed. Make a note of the time and flag it as valid.
                e_idx = this_line.find('Exit')
                game.time = totime(this_line[0:e_idx])
                valid_game = True
                
                n += 1


            elif this_line.find(' ShutdownGame:') > 0:
                break


            j += 1
            
        i = j + 1
        
        
        if valid_game == True:
            server.time = server.time + game.time - min(game.ptime.values())
            cgames.append(game)                 # Append game to list of games
            valid_game = False                  # and start it over
            
    else:
    # Keep looking until the beginning of a game is found
        i += 1
        



def csum(A):
    """Column-wise addition of lists. Returns a list."""
    # Check whether A is multidimensional
    if type(A[0]) is not list and type(A[0]) is not tuple:
        S = A
        return S
    # Check that A is not jagged
    for row in A:
        if len(row) != len(A[0]):
            print "Not a matrix."
    # Do the damn summation cause Python can't be bothered to do it alone
    S = []
    for i in range(len(A[0])):
        S.append(sum([row[i] for row in A]))
    return S



def player_stats(game_number, player_name):
    """Gather the relevant numbers on a per-game, per-player basis."""
    game = cgames[game_number]
    
    if player_name not in game.validp:
    # Check player has actually played game and is tagged as valid, gtfo if it hasn't
        return 0
        
    time   = game.time - game.ptime[player_name]
    ping   = game.players[player_name][0]
    hand   = game.handicap[player_name]

    if (game.gametype != '4') and (game.gametype != '3'):
        if game.players[player_name][1] == 1:
            win = 1
        else:
            win = 0
    else:
        try:             # We 'try' it to avoid problems with spectators
            if game.ctfscores[int(game.teams[player_name]) - 1] == max(game.ctfscores):
                win = 1
            else:
                win = 0
        except(IndexError):
            win = 0
        
    awards = []
    awards.extend([n[1] for n in sorted(game.awards[player_name].iteritems())])
    
    wfrags = [n for n in game.killsp['<world>']].count(player_name)
    deaths = game.deathsp[player_name]
    suics  = len([n for n in game.killsp[player_name]])
    frags  = sum( game.weapons[player_name].values() )
    
    weapons = [n[1] for n in game.killsp[player_name] if n[0] != player_name]
    weapon_count = []                                         # per weapon frags
       
    wlist = ['SHOTGUN', 'GAUNTLET', 'MACHINEGUN', 'GRENADE', 'GRENADE_SPLASH',\
            'ROCKET', 'ROCKET_SPLASH', 'PLASMA', 'PLASMA_SPLASH', 'RAILGUN',  \
            'LIGHTNING', 'BFG10K', 'BFG10K_SPLASH', 'TELEFRAG', 'NAIL', 'CHAIN']
    for w in wlist:
        weapon_count.append(game.weapons[player_name][w[0:3]])

    if game.gametype == '4':
        flags_taken = game.ctf[player_name]['0']
        #flags_captd = game.ctf[player_name].count('1')       # equal to cap award
        flags_retrd = game.ctf[player_name]['2']
        flag_fraggd = game.ctf[player_name]['3']
        ctf_events  = (flags_taken, flags_retrd, flag_fraggd)
    else:
        ctf_events = (0, 0, 0)
    
    #armor = game.itemsp[player_name].count('item_armor_combat')
    #mega  = game.itemsp[player_name].count('item_health_mega')
    #quad  = game.itemsp[player_name].count('item_quad')
    #regen = game.itemsp[player_name].count('item_regen')
    #haste = game.itemsp[player_name].count('item_haste')
    #items = [armor, mega, quad, regen, haste]

    key = ['win', 'time', 'handicap', 'ping', 'frags', 'deaths', 'suics', 'wfrags', 'awards', 'weapon count', 'ctf_events']#, 'items']

    return [key, win, time, hand, ping, frags, deaths, suics, wfrags, awards, weapon_count, ctf_events]#/map, items]



def allnames(cgames):
    """Return names of all valid players in log."""
    allnames = set()
    for game in cgames:
        allnames.update(game.validp)
    return allnames



def player_stats_total(cgames):
    """Get accumulated stats per player for all players in log.""" 
    all_players = []
    for name in allnames(cgames):
        win, time = 0, 0
        hand, ping, weapon_count, ctf_events = [], [], [], []
        frags, deaths, suics, wfrags = 0, 0, 0, 0
        awards_a, awards_c, awards_d, awards_e, awards_i = 0, 0, 0, 0, 0
        items = []
        for i in xrange(len(cgames)):
            if name not in cgames[i].validp:         # ignore no valid players
                pass
            else:
                game_stats = player_stats(i, name)
                win      = win      + game_stats[1]
                time     = time     + game_stats[2]
                hand.append(int(game_stats[3]))                
                ping.append(int(game_stats[4]))
                frags    = frags    + game_stats[5]
                deaths   = deaths   + game_stats[6]
                suics    = suics    + game_stats[7]
                wfrags   = wfrags   + game_stats[8]
                awards_a = awards_a + game_stats[9][0]
                awards_c = awards_c + game_stats[9][1]
                awards_d = awards_d + game_stats[9][2]
                awards_e = awards_e + game_stats[9][3]
                awards_i = awards_i + game_stats[9][4]
                weapon_count.append(game_stats[10])
                ctf_events.append(game_stats[11])
                #items.append(game_stats[12])
        if frags == 0:  # Take rid of those pesky players unable to set autodownload 'on'
            pass        # This isn't quite right, there might actually be players this bad
        else:
            one_player = {'name':name,  'games':len(hand),  'won':win,                      \
                          'time':time,  'hand':sum(hand)/len(hand),                         \
                          'ping': [min(ping), sum(ping)/len(ping), max(ping)],              \
                          'frags':frags, 'deaths':deaths, 'suics':suics, 'wfrags':wfrags,   \
                          'excellent':awards_e,  'impressive':awards_i,                     \
                          'defence':awards_d,   'capture':awards_c,  'assist':awards_a,     \
                          'weapons':csum(weapon_count), 'ctf':csum(ctf_events)              }
                          #'items':csum(items)                                              }
            all_players.append(one_player)
    return all_players



def name_colour(nick):
    '''Parse Quake colour codes to HTML (uses pyqscores' CSS stylesheet).
       This will also strip any '^' symbol contained in the nick. Oh, well.
       Also, colour code 8 ignored.'''
    limit = len(nick) -1
    if nick.rfind('^') > -1 and nick.rfind('^') < limit:
        idx = nick.find('^')
        if nick[idx + 1] == '0':
            name = nick[0:idx] + '<SPAN class="c0">' + nick[idx+2:] + '</SPAN>'
        elif nick[idx + 1] == '1':
            name = nick[0:idx] + '<SPAN class="c1">' + nick[idx+2:] + '</SPAN>'
        elif nick[idx + 1] == '2':
            name = nick[0:idx] + '<SPAN class="c2">' + nick[idx+2:] + '</SPAN>'
        elif nick[idx + 1] == '3':
            name = nick[0:idx] + '<SPAN class="c3">' + nick[idx+2:] + '</SPAN>'
        elif nick[idx + 1] == '4':
            name = nick[0:idx] + '<SPAN class="c4">' + nick[idx+2:] + '</SPAN>'
        elif nick[idx + 1] == '5':
            name = nick[0:idx] + '<SPAN class="c5">' + nick[idx+2:] + '</SPAN>'
        elif nick[idx + 1] == '6':
            name = nick[0:idx] + '<SPAN class="c6">' + nick[idx+2:] + '</SPAN>'
        elif nick[idx + 1] == '7':
            name = nick[0:idx] + '<SPAN class="c7">' + nick[idx+2:] + '</SPAN>'
    	else:
            name = nick[0:idx] + nick[idx+1:]
    else:
        name = nick
    if name.find('^') > -1 and nick.rfind('^') < limit:
        name = name_colour(name)
    return name



def is_number(s):
    '''Is 's' a number?'''
    try:
        int(s)
        return True
    except ValueError:
        return False
    
   
    
def results_ordered(R, option, maxnumber):
    '''Sort the dictionary-storing list R according to the key specified
    by option. The inexistent keys 'frag_death_ratio' and 'won_percentage'
    are added here for convenience. maxnumber limits the size of the output.'''
    if is_number(maxnumber) is False:
        print "\nINVALID MAXNUMBER VALUE IN FUNCTION results_ordered. Check MAXPLAYERS option.\n"
        return
    if maxnumber > len(R):
        maxnumber = len(R)
    elif maxnumber <= 0:
        print "\nINVALID MAXNUMBER VALUE IN FUNCTION results_ordered. Check MAXPLAYERS option.\n"
        return
    if option == 'frag_death_ratio':
        Rordered = sorted( R, key = lambda dic: float(dic['frags'])/dic['deaths'], reverse=True )
    elif option == 'won_percentage':
        Rordered = sorted( R, key = lambda dic: float(dic['won'])/dic['games'], reverse=True )
    elif option == 'frags_per_hour':
        Rordered = sorted( R, key = lambda dic: float(dic['frags'])/dic['time'], reverse=True )
    elif option == 'name':
        Rordered = sorted( R, key = lambda dic: float(dic['frags'])/dic['deaths'], reverse=False )
    elif option in R[0].keys():
        Rordered = sorted( R, key = lambda dic: dic[option], reverse=True )
    elif option not in R[0].keys():
        print "\nINVALID ORDERING OPTION IN FUNCTION results_ordered(). Check spelling?\n"
        return
    return Rordered[0:maxnumber]



def make_table(L,  style1_even, style1_odd, style2_even, style2_odd):
    '''Generate HTML table from input data stored in list L.
       style1_even and style1_odd is the style class for the even and odd entries
       of the first column (player names).
       style2_even and style2_odd is the equivalent for the rest of entries.'''
    lines = []
    
    for i in xrange(len(L)):
        if (mod(i,2) == 0):
            string1 = '<TR>\n'
            string1 = string1 + '<TD><DIV class="%s">' %style1_even + str(L[i][0]) +'\n'
            string1 = string1 + '</DIV></TD>'
            string2 = ''
        else:
            string1 = '<TR>\n'
            string1 = string1 + '<TD><DIV class="%s">' %style1_odd + str(L[i][0]) +'\n'
            string1 = string1 + '</DIV></TD>'
            string2 = ''
        for j in xrange(1,len(L[0])):
            if (mod(i,2) == 0):
                string2 =  string2 + '<TD><DIV class="%s">' %style2_even + str(L[i][j]) + '\n'
                string2 = string2 + '</DIV></TD>\n'
            else:
                string2 =  string2 + '<TD><DIV class="%s">' %style2_odd + str(L[i][j]) + '\n'
                string2 = string2 + '</DIV></TD>\n'
        string2 = string2 + '</TR>\n'
        lines.append(string1)
        lines.append(string2)
    lines.append('</TABLE>')
    return lines



def write_table(table_header, table_data, style1_even,\
                style1_odd, style2_even, style2_odd):
    '''Write HTML table to file from a 'header' (The first part of the table
       up to the actual data) and the data stored in a list.'''
    f.write(table_header)
    lines = make_table(table_data, style1_even, style1_odd,\
                        style2_even, style2_odd)
    for line in lines:
        print >> f, line



def get_quotes(cgames):
    '''Create non repeating list of quotes from list of games.'''
    if len(cgames) != 0:
        quotes_list = [cgames[i].quotes for i in xrange(len(cgames)) if cgames[i].quotes != []]
        quotes_list = list(set([item for sublist in quotes_list for item in sublist]))
    else:
        quotes_list = []
    return quotes_list

quotes_list = get_quotes(cgames)


# Deal with cache files, whether present or absent
if cache_present == 'yes':
    server_old = cache[-3]               # Make a copy of server data in cache
    quotes_list.extend(cache[-4])        # Add previous quotes to current list
    quotes_list = list(set(quotes_list)) # Eliminate duplicates
    del cache[-4:]                       # Get rid of extra bits in cache file
    R_old = cache                        # Make a copy of previous player data

    if len(cgames) != 0:
    # If there are new games in log, add new data to that from the cache
        R = player_stats_total(cgames)   # Process new player data found in log
        server_old.frags = server_old.frags + server.frags
        server_old.gtype = server.gtype  # This we don't add, we just update it
        server_old.time  = server_old.time + server.time
        old_names = [player['name'] for player in R_old]
        keys = R[0].keys()
        keys.remove('name')
        keys.remove('ping')
        keys.remove('weapons')
        keys.remove('ctf')
        keys.remove('hand')
        for player in R:
            if player['name'] in old_names:
                index = old_names.index(player['name'])
                min_ping = min( R_old[index]['ping'][0], player['ping'][0] )
                max_ping = max( R_old[index]['ping'][2], player['ping'][2] )
                avg_png1 = R_old[index]['ping'][1]
                avg_png2 = player['ping'][1]
                games1   = R_old[index]['games']
                games2 = player['games']
                ave_ping = (avg_png1*games1 + avg_png2*games2)/(games1 + games2)
                R_old[index]['ping'] = [min_ping, ave_ping, max_ping]
                R_old[index]['weapons'] = csum([ R_old[index]['weapons'], player['weapons'] ])
                R_old[index]['ctf']  = csum([ R_old[index]['ctf'], player['ctf'] ] )
                avg_han1 = R_old[index]['hand']
                avg_han2 = player['hand']
                R_old[index]['hand'] = (avg_han1*games1 + avg_han2*games2)/(games1+games2)
                for key in keys:
                    R_old[index][key] = csum( [ [R_old[index][key]], [player[key]] ] )[0]
            else:
                R_old.append(player)
        R = R_old
        server = server_old
    else:
        R = R_old                   # If there's no new data, just use the cache
        server = server_old         # Ditto for server data
else:
    if len(cgames) != 0:            # No cache: process all the games
        R = player_stats_total(cgames)
    else:                           # Unless there are none
        print '\nNo valid games found in log. Play a bit more.\n'
        quit()
    



# This is the place to dump the updated cache, once some extra
# info is added at the end for future reference
cache = R
log_size = os.path.getsize(sys.argv[1])
date_now = datetime.now().strftime("%c")
cache.append(quotes_list)
cache.append(server)
cache.append(('lines read',k-1))
cache.append(('Log size on ' + date_now, log_size))
cPickle.dump(cache, open(cache_file, 'wb'))
del cache[-4:]  # Once written delete the extra bits to avoid problems below



# Stats only tested with game types 0 and 4, but we'll
# report the correct game type in any case.
gametypes = {0: 'Death Match', 1: '1 vs 1', 2: 'Single Death Match',               \
             3: 'Team Death Match', 4: 'Capture the Flag', 5: 'One-Flag CTF',      \
             6: 'Overload', 7: 'Harvester', 8: 'Elimination', 9: 'CTF Elimination',\
            10: 'Last Man Standing', 11: 'Double Elimination', 12: 'Domination' }

# If user specifies game type, report it regardless of what pyqscore parsed
if GTYPE_OVERRIDE in '':
    server.gtype = gametypes[server.gtype]
elif GTYPE_OVERRIDE in ['ctf', 'CTF']:
    server.gtype = 'Capture the Flag'
elif GTYPE_OVERRIDE in ['dm', 'DM']:
    server.gtype = 'Death Match'
else:
    server.gtype = 'Unknown'

R = results_ordered(R,SORT_OPTION,MAXPLAYERS)        # Sort stuff out to taste

# Dump data in JSON format if so required. Do this now, once data is sorted
# but before parsing the colour codes: they aren't useful without the .css
if DUMP_DATA in ('yes', 'Yes', 'YES'):
    dump_file = log_file[:-4] + '_dump.json'
    f = open(dump_file, 'w')
    json.dump(R, f, sort_keys = True)
    f.close()
    


def apply_ban(R, BAN_LIST):
    ''''Possibly naive implementation of a black list of players.'''
    R_names = [player['name'] for player in R]
    ban_list_index = []

    for name in BAN_LIST:
        if name in R_names:
            ban_list_index.append(R_names.index(name))
    for i in sorted(ban_list_index, reverse = True):
        del(R[i])
    return R
    
R = apply_ban(R, BAN_LIST)


for player in R:
    player['name'] = name_colour(player['name'])  # Coloured nicks are cooler




# Time to put data together for the different tables. The entries
# of each table are stored in a list containing data derived from R.

main_table_data = []

for player in R:
    main_table_data.append([ player['name'],    player['games'],    player['won'],      \
                             str(timedelta(seconds=player['time'])),                    \
                             player['hand'],    player['ping'][0],  player['ping'][1],  \
                             player['ping'][2], player['frags'],    player['deaths'],   \
                             player['suics'],   player['wfrags'],   player['excellent'],\
                             player['impressive']  ])


weapons_table_data = []                    # List storing data for weapons table

for i in xrange(len(R)):
    weapons_table_data.append([R[i]['name']])
    weapons_table_data[i].extend(R[i]['weapons'][0:3])        # SHOTG, GAUNT, MACHG
    weapons_table_data[i].append(sum(R[i]['weapons'][3:5]))   # GRENADE
    weapons_table_data[i].append(sum(R[i]['weapons'][5:7]))   # ROCKET
    weapons_table_data[i].append(sum(R[i]['weapons'][7:9]))   # PLASMA
    weapons_table_data[i].extend(R[i]['weapons'][9:11])       # RAIL, LIGHTNING
    weapons_table_data[i].extend(R[i]['weapons'][14:])        # NAILGUN, CHAINGUN
    weapons_table_data[i].append(sum(R[i]['weapons'][11:13])) # BFG
    weapons_table_data[i].extend(R[i]['weapons'][13:14])      # TELEFRAG


for i in xrange(len(weapons_table_data)):  # Cut down the decimals and work %s out
    for j in xrange(1, len(weapons_table_data[i])):
        weapons_table_data[i][j] = str(round((100.*weapons_table_data[i][j]/R[i]['frags']), 2))


stats_table_data = []                      # Another table with more numbers
for i in xrange(len(R)):
    #name           #% games won    #frags/deaths    #frags/hour        #frags/game    
    #deaths/hour    #deaths/game    #suic+fall/hour  #suic+fall/game    #efficiency
    stats_table_data.append([       R[i]['name'],                                  \
                             100.*  R[i]['won']    / R[i]['games'],                \
                             1.*    R[i]['frags']  /(1+R[i]['deaths']),            \
                             3600.* R[i]['frags']  / R[i]['time'],                 \
                             1.*    R[i]['frags']  / R[i]['games'],                \
                             3600.* R[i]['deaths'] / R[i]['time'],                 \
                             1.*    R[i]['deaths'] / R[i]['games'],                \
                             3600.*(R[i]['suics']  + R[i]['wfrags'])/R[i]['time'], \
                             1.*(   R[i]['suics']  + R[i]['wfrags'])/R[i]['games'],\
                             100.*  R[i]['frags']  /(1+R[i]['frags']+R[i]['deaths'])])

for i in xrange(len(stats_table_data)):
    for j in xrange(1, len(stats_table_data[i])):
        stats_table_data[i][j] = str(round(stats_table_data[i][j], 2))


quotes_table = []                           # Random quotes table
if len(quotes_list) > 0:
    for i in xrange(NUMBER_OF_QUOTES):
        a = quotes_list[int(randint(0,len(quotes_list)-1))]
        quotes_table.append([ name_colour(a[0]), a[1] ] )


ctf_table = []                              # Table with CTF-related numbers
for i in xrange(len(R)):
    ctf_table.append( [ R[i]['name'] ] )
    ctf_table[i].extend( R[i]['ctf'] )
    ctf_table[i].extend([ R[i]['defence'], R[i]['assist'], R[i]['capture'] ])



# Everything is in place to start writing the HTML output file.
# Some raw strings needed to do so follow:


html_header = r'''
<!DOCTYPE html PUBLIC "-//W3C//DTD HTML 4.01//EN"
   "http://www.w3.org/TR/html4/strict.dtd">
      
<HTML>
<HEAD>
<META http-equiv="Content-Type" content="text/html; charset=utf-8">
<META http-equiv="Content-Language" content="en">
<META name="description" content="pyqscore OpenArena stats">
<META name="keywords" content="stats">
<META name="robots" content="noindex, nofollow">
<LINK rel="icon" href="favicon.ico" type="image/x-icon">
<LINK rel="stylesheet" href="pyqscore_style.css" type="text/css">
<TITLE>OpenArena stats 24h</TITLE>
</HEAD>

<BODY>
<DIV id="marco">


<DIV class="cuadropresentacion">
<DIV class="update">Last updated on %s </DIV>
<DIV class="intro">OpenArena Stats 24h</DIV>


<DIV class="centrartabla">
<TABLE class="tablaserver">

<TR>
<TD><DIV class="server">Server Name:</DIV></TD>
<TD><DIV class="datoserver">%s </DIV></TD>
<TD><DIV class="server">Server Total Time:</DIV></TD>
<TD><DIV class="datoserver">%s </DIV></TD>
</TR>

<TR>
<TD><DIV class="server">Game Type:</DIV></TD>
<TD><DIV class="datoserver">%s </DIV></TD>
<TD><DIV class="server">Total Frags:</DIV></TD>
<TD><DIV class="datoserver">%s </DIV></TD>
</TR>

<TR>
<TD colspan="2"><DIV class="server"><a href="../../stats.htm" title="back"><< back to statistics</a></DIV></TD>
<TD colspan="2"><DIV class="server"><a href="monthly.htm" title="current month">current month >></a></DIV></TD>
</TR>


</TABLE>
</DIV>
</DIV>

<DIV class="cuadro">
<DIV class="titulocuadro"></DIV>
'''

main_table_header = r'''
<TABLE class="tabladatos">
<TR>
<TH><DIV class="tituloup"></DIV></TH>
<TH><DIV class="tituloup">Total</DIV></TH>
<TH><DIV class="tituloup">Games</DIV></TH>
<TH><DIV class="tituloup">Time</DIV></TH>
<TH><DIV class="tituloup">Mean</DIV></TH>
<TH><DIV class="tituloup">Lowest</DIV></TH>
<TH><DIV class="tituloup">Average</DIV></TH>
<TH><DIV class="tituloup">Highest</DIV></TH>
<TH><DIV class="tituloup">Total</DIV></TH>
<TH><DIV class="tituloup">Total</DIV></TH>
<TH><DIV class="tituloup">Total</DIV></TH>
<TH><DIV class="tituloup">Deads</DIV></TH>
<TH><DIV class="tituloup">Excellent</DIV></TH>
<TH><DIV class="tituloup">Impressive</DIV></TH>
</TR>

<TR>
<TH><DIV class="tituloup3"></DIV></TH>
<TH><DIV class="tituloup3">games</DIV></TH>
<TH><DIV class="tituloup3">won</DIV></TH>
<TH><DIV class="tituloup3">played</DIV></TH>
<TH><DIV class="tituloup3">handicap</DIV></TH>
<TH><DIV class="tituloup3">ping</DIV></TH>
<TH><DIV class="tituloup3">ping</DIV></TH>
<TH><DIV class="tituloup3">ping</DIV></TH>
<TH><DIV class="tituloup3">frags</DIV></TH>
<TH><DIV class="tituloup3">deaths</DIV></TH>
<TH><DIV class="tituloup3">suicides</DIV></TH>
<TH><DIV class="tituloup3">falling</DIV></TH>
<TH><DIV class="tituloup3">awards</DIV></TH>
<TH><DIV class="tituloup3">awards</DIV></TH>
</TR>
'''

ctf_table_header = r'''
<DIV class="centrartabla2">
<TABLE class="tabladatos">

<TR>
<TH><DIV class="tituloup"></DIV></TH>
<TH><DIV class="tituloup">Flags</DIV></TH>
<TH><DIV class="tituloup">Flags</DIV></TH>
<TH><DIV class="tituloup">Flagcarrier</DIV></TH>
<TH><DIV class="tituloup">Defence</DIV></TH>
<TH><DIV class="tituloup">Assist</DIV></TH>
<TH><DIV class="tituloup">Capture</DIV></TH>
<TR>

<TR>
<TD><DIV class="tituloup3"></DIV></TD>
<TD><DIV class="tituloup3">taken</DIV></TD>
<TD><DIV class="tituloup3">returned</DIV></TD>
<TD><DIV class="tituloup3">frags</DIV></TD>
<TD><DIV class="tituloup3">awards</DIV></TD>
<TD><DIV class="tituloup3">awards</DIV></TD>
<TD><DIV class="tituloup3">awards</DIV></TD>
</TR>
'''

stats_table_header = r'''
<DIV class="centrartabla2">
<TABLE class="tabladatos">

<TR>
<TH><DIV class="tituloup"></DIV></TH>
<TH><DIV class="tituloup">Games</DIV></TH>
<TH><DIV class="tituloup">Frags/</DIV></TH>
<TH><DIV class="tituloup">Frags</DIV></TH>
<TH><DIV class="tituloup">Frags</DIV></TH>
<TH><DIV class="tituloup">Deaths</DIV></TH>
<TH><DIV class="tituloup">Deaths</DIV></TH>
<TH><DIV class="tituloup">Suics+falling</DIV></TH>
<TH><DIV class="tituloup">Suics+falling</DIV></TH>
<TH><DIV class="tituloup">Efficiency</DIV></TH>
</TR>

<TR>
<TH><DIV class="tituloup3"></DIV></TH>
<TH><DIV class="tituloup3">won %</DIV></TH>
<TH><DIV class="tituloup3">deaths</DIV></TH>
<TH><DIV class="tituloup3">per hour</DIV></TH>
<TH><DIV class="tituloup3">per game</DIV></TH>
<TH><DIV class="tituloup3">per hour</DIV></TH>
<TH><DIV class="tituloup3">per game</DIV></TH>
<TH><DIV class="tituloup3">per hour</DIV></TH>
<TH><DIV class="tituloup3">per game</DIV></TH>
<TH><DIV class="tituloup3"></DIV></TH>
</TR>
'''

weapon_table_header = r'''
<DIV class="centrartabla2">
<TABLE class="tabladatos">

<TR>
<TH><DIV class="tituloup4"></DIV></TH>
<TH><DIV class="tituloup4">Shotgun</DIV></TH>
<TH><DIV class="tituloup4">Gauntlet</DIV></TH>
<TH><DIV class="tituloup4">Machinegun</DIV></TH>
<TH><DIV class="tituloup4">Grenade</DIV></TH>
<TH><DIV class="tituloup4">Rocket</DIV></TH>
<TH><DIV class="tituloup4">Plasma</DIV></TH>
<TH><DIV class="tituloup4">Railgun</DIV></TH>
<TH><DIV class="tituloup4">Lightning</DIV></TH>
<TH><DIV class="tituloup4">Nailgun</DIV></TH>
<TH><DIV class="tituloup4">Chaingun</DIV></TH>
<TH><DIV class="tituloup4">BFG</DIV></TH>
<TH><DIV class="tituloup4">Telefrag</DIV></TH>
</TR>

<TR>
<TH><DIV></DIV></TH>
<TH><DIV class="icons"> <IMG SRC="icons/shotgun.png" width="20" height="20" alt=""></DIV></TH>
<TH><DIV class="icons"> <IMG SRC="icons/gauntlet.png" width="20" height="20" alt=""></DIV></TH>
<TH><DIV class="icons"> <IMG SRC="icons/machinegun.png" width="20" height="20" alt=""></DIV></TH>
<TH><DIV class="icons"> <IMG SRC="icons/grenade.png" width="20" height="20" alt=""></DIV></TH>
<TH><DIV class="icons"> <IMG SRC="icons/rocket.png" width="20" height="20" alt=""></DIV></TH>
<TH><DIV class="icons"> <IMG SRC="icons/plasma.png" width="20" height="20" alt=""></DIV></TH>
<TH><DIV class="icons"> <IMG SRC="icons/railgun.png" width="20" height="20" alt=""></DIV></TH>
<TH><DIV class="icons"> <IMG SRC="icons/lightning.png" width="20" height="20" alt=""></DIV></TH>
<TH><DIV class="icons"> <IMG SRC="icons/nailgun.png" width="20" height="20" alt=""></DIV></TH>
<TH><DIV class="icons"> <IMG SRC="icons/chaingun.png" width="20" height="20" alt=""></DIV></TH>
<TH><DIV class="icons"> <IMG SRC="icons/bfg.png" width="20" height="20" alt=""></DIV></TH>
<TH><DIV class="icons"> <IMG SRC="icons/teleporter.png" width="20" height="20" alt=""></DIV></TH>
</TR>
'''

items_table_header = r'''
<TABLE class="tabladatos">

<TR>
<TH><DIV class="tituloup"></Div></TD>
<TH><DIV class="tituloup">Awards</DIV></TD>
'''



#html_file = str(sys.argv[1][:-3]) + 'html'
html_file = 'daily.htm'

f = open(html_file, 'w')
#print >> f, html_header .format(today=datetime.now().strftime("%c"))
print >> f, html_header %(datetime.now().strftime("%c"), name_colour(server.hostname),\
                      str(timedelta(seconds=server.time)), server.gtype, server.frags )


    
write_table(main_table_header, main_table_data, 'jugador', 'jugador2',\
                                                       'dato', 'dato2')
            

# Pyqscore doesn't separate logs from different game types
# If the log contains a mix of CTF and non-CTF games, the
# only effort made to tell what's going on is to check whether
# player at position R[0] has non-zero ctf events.
if ((R[0]['ctf'] != [0, 0, 0]) or GTYPE_OVERRIDE in ('ctf','CTF')) \
		              and (GTYPE_OVERRIDE) not in ('dm','DM'):
    write_table(ctf_table_header, ctf_table, 'jugador', 'jugador2', \
                                                     'dato', 'dato2')
    print >> f, '</DIV>'

write_table(stats_table_header, stats_table_data, 'jugador', 'jugador2',\
                                                         'dato', 'dato2')
print >> f, '</DIV>'

write_table(weapon_table_header, weapons_table_data, 'jugador2', 'jugador',\
                                                            'dato2', 'dato')
print >> f, '</DIV>'

print >> f, r'''</DIV>
<DIV class="endnote">Powered by pyqscore</DIV>
</DIV>
</BODY>
</HTML>'''

f.close()


try:    # Why do I use a try statement? I don't even remember...
    if OPEN_BROWSER in ('yes', 'Yes', 'YES'):
        webbrowser.open_new(html_file)
        #os.system('xdg-open %s' %html_file)
except:
    pass



##### read file line by line ####
#file_in = open(sys.argv[1],'r')

#line = file_in.readline()

#while line:
    #### do stuff ####
    #print line
    #line = file_in.readline()
